/*
 * Copyright 2002
 *  Bob Shaffer II
 *  4517 West Melissa Court
 *  Jacksonville, Florida 32210
 *  United States of America
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 */

#include "MySQL_Database.h"
#include <sstream>

MySQL_Database::MySQL_Database(const char *database,
                               const char *user,
                               const char *password,
                               const char *host,
                               int port) {
    // Constructor
    mysql_init(&DB);
    mysql_real_connect(&DB, host, user, password, database, port, NULL, 0);
}

MySQL_Database::~MySQL_Database() {
    // Destructor
    mysql_close(&DB);
}

void MySQL_Database::Direct_Query(Table &Result, string query, Delimited_Row *FieldList) {
    // Query database for a read-only table
    Result.clear();
    mysql_real_query(&DB, query.c_str(), query.length());

    // If our query was not a "Select", return now.
    for (int c=0; c<query.length(); ++c) query[c]=tolower(query[c]);
    if (query.find("select")==query.npos) return;

    MYSQL_RES *result(mysql_use_result(&DB));
    MYSQL_ROW current_row( mysql_fetch_row(result) );
    int TableWidth( mysql_num_fields(result) );
    if (FieldList) {
        FieldList->clear();
        MYSQL_FIELD *fieldnames( mysql_fetch_fields(result) );
        for (int FNum=0; FNum<TableWidth; ++FNum) {
            FieldList->cell(FNum)=fieldnames[FNum].name;
        }
    }
    for (;;) {
        if (!current_row) break;
        Delimited_Row NewRow;
        NewRow.set_delimiter(Result.get_delimiter());
        for (int fieldnum=0; fieldnum<TableWidth; ++fieldnum) {
            if (!current_row[fieldnum]) {
                NewRow.cell(fieldnum)="";
            } else {
                NewRow.cell(fieldnum)=current_row[fieldnum];
            }
        }
        Result.push_back(NewRow);
        current_row = mysql_fetch_row(result);
    }
    mysql_free_result(result);
}

void MySQL_Database::Direct_Query(string query) {
    mysql_real_query(&DB, query.c_str(), query.length());
}

void MySQL_Database::Query(MySQL_Table &Result,
                      const char *table,
                      const char *fields,
                      const char *params) {
    // Query database for an updatable table
    long f;
    string FieldList(fields);
    // I had to find a reliable way to check for this damn ID field
    Comma_Delimited_Row TempFieldRow;
    TempFieldRow.set_row(FieldList);
    bool foundID=false;
    // For each field name in the row
    for (int f=0; f<TempFieldRow.length() && (!foundID); ++f) {
        // remove all spaces
        int c=0;            
        while (c<TempFieldRow[f].length()) {
            if (isspace(TempFieldRow[f][c])) TempFieldRow[f].erase(c);
            else ++c;
        }
        if (TempFieldRow[f]=="ID" || TempFieldRow[f]=="*") foundID=true;
    }
    if (!foundID) {
        // add the ID field to the end to avoid any confusion
        FieldList+=", ID";
    }
        
    Result.TableName = table;
    Result.mDB=this;

    ostringstream qstream;
    qstream<<"select "<<FieldList<<" from "<<table;
    if (params) qstream<<' '<<params;
    Direct_Query(Result.Data, qstream.str(), &(Result.Fields));
    Result.Original.clear();
    Result.Original=Result.Data;
}

int MySQL_Database::errornum() {
    return mysql_errno(&DB); 
}

string MySQL_Database::Escape_String(string s) {
    char temp[s.length()*2+1];
    mysql_real_escape_string(&DB, temp, s.c_str(), s.length());
    return string(temp);
}

MySQL_Table::MySQL_Table() {
    mDB=NULL;
}

MySQL_Table::MySQL_Table(MySQL_Database &DBRef,
                         const char *table,
                         const char *fields,
                         const char *params) {
    DBRef.Query(*this, table, fields, params);
}


MySQL_Table::~MySQL_Table() { Update(); }

string &MySQL_Table::cell(long row, long col) {
    return Data.cell(row, col);
}

string &MySQL_Table::cell(long row, string cellName) {
    return cell(row, FieldIndex(cellName));
}

Table &MySQL_Table::GetTable() {
    return Data;
}

long MySQL_Table::FieldIndex(string FieldName) {
    long Indexn=0;
    while (Indexn<Fields.length()) {
        if (Fields.cell(Indexn)==FieldName) return Indexn;
        ++Indexn;
    }
    return -1;
}

long MySQL_Table::ID_Index() { return FieldIndex("ID"); }

my_ulonglong MySQL_Table::ID(long row) {
    if (ID_Index()==-1) return 0;
    else return atol(cell(row, ID_Index()).c_str());
}

long MySQL_Table::length() {
    return Data.length();
}

long MySQL_Table::Insert(Delimited_Row &NewRow) {

//  This method should be working perfectly fine

    ostringstream insert_qstream, id_qstream, select_qstream;
    insert_qstream<<"insert into "<<TableName<<" set ";
    bool firstField( true );
    for (long f=0; f<Fields.length(); ++f) {
        if (f!=ID_Index()) {
            char *TempStr=new char[NewRow[f].length()*2+1];
            mysql_real_escape_string(&(mDB->DB), TempStr,
                            NewRow[f].c_str(), NewRow[f].length());
            if (firstField) firstField=false;
            else insert_qstream<<", ";
            insert_qstream<<Fields.cell(f)<<"=\'"<<TempStr<<'\'';
            delete [] TempStr;
        }
    }

//    cerr<<"Executing Query ->"<<insert_qstream.str()<<"<-\n";

    mDB->Direct_Query(insert_qstream.str());
    ostringstream Big_ID_Num;

//    cerr<<"Query done.  Retrieving ID for new row...";

    Big_ID_Num<<mysql_insert_id(&(mDB->DB));

//    cerr<<"<" << Big_ID_Num.str() << "> done.\n";

    NewRow.cell(ID_Index())=Big_ID_Num.str();
    Data.push_back(NewRow);
    Original.push_back(NewRow);
    long RowNum( Data.length()-1 );
    return RowNum;
}

void MySQL_Table::Remove(long row) {
    ostringstream qstream;
    qstream<<"delete from "<<TableName<<" where ID="<<ID(row)<<';';
    string query(qstream.str());
    mysql_real_query(&(mDB->DB), query.c_str(), query.length());
    Original.remrow(row);
    Data.remrow(row);
}

void MySQL_Table::Update() {
    // if (Data.length()!=Original.length()) return; // unexpected (necessary?)
    for (long a=0; a<Data.length(); ++a) {
        string UpdateList;
        // if (Data[a].length()!=Original[a].length()) return; // unexpected (necessary?)
        for (long b=0; b<Data[a].length(); ++b) {
            if (b==ID_Index()) continue;
            if (Data[a][b]!=Original[a][b]) {
                char *TempStr=new char[Data[a][b].length()*2+1];
                mysql_real_escape_string(&(mDB->DB), TempStr,
                                Data[a][b].c_str(), Data[a][b].length());
                if (UpdateList.length()) UpdateList+=", ";
                UpdateList+=Fields[b]+"=\'"+TempStr+"\'";
                delete [] TempStr;
            }
        }
        if (UpdateList.length()) {
            ostringstream qstream;
            qstream<<"update "<<TableName
                   <<" set "<<UpdateList
                   <<" where ID="<<ID(a)<<';';
            string query(qstream.str());
            mysql_real_query(&(mDB->DB), query.c_str(), query.length());
            Original[a]=Data[a];
        }
    }
}

